/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2007 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.deployment;

import java.util.List;
import java.util.logging.Level;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.facade.Deployment;
import org.ow2.orchestra.facade.def.ProcessDefinition;
import org.ow2.orchestra.facade.def.full.ProcessFullDefinition;
import org.ow2.orchestra.facade.exception.DeploymentException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.wsdl.WsdlDefinitions;
import org.ow2.orchestra.parsing.BpelParser;
import org.ow2.orchestra.parsing.binding.DefinitionKeeper;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.internal.xml.Parse;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.handlers.UndeployedProcessHandler;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.Namespaces;
import org.ow2.orchestra.util.ProcessResourcesRepository;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.xml.sax.InputSource;

/**
 * takes {@link Deployment}s as input, parses the contents into a
 * {@link BpelProcess}, stores the {@link BpelProcess} into the
 * {@link Repository} and publishes the services with a {@link Publisher}.
 *
 * <p>
 * processRepository and servicePublisher have to be injected by the user (or an
 * IoC framework) before the deploy method is invoked.
 * </p>
 *
 * <p>
 * This class is thread safe and can be shared between multiple threads.
 * </p>
 *
 * @author Tom Baeyens, Goulven Le Jeune
 */
public class Deployer {

  /** parser to use to parse wsdl files. */
  //private BpelWsdlParser bpelWsdlParser = null;

  /** parser to use to parse bpel files. */
  private BpelParser bpelParser = null;

  /**
   * Default constructor.
   */
  public Deployer() {
    //bpelWsdlParser = new BpelWsdlParser();
    this.bpelParser = new BpelParser();
  }

  /**
   * Thanks to the deployment object given in parameter, parse all wsdl files,
   * bpel file. Then store the process in processRepository and publish bpel
   * process services.
   *
   * @param bpelDeployment
   *            populated bpelDeployment
   * @param environment
   *            environment
   * @return created BpelProcess
   */
  public ProcessFullDefinition deploy(final Deployment bpelDeployment, final Environment environment) {
    Misc.checkArgsNotNull(bpelDeployment, environment);

    Misc.log(Level.INFO, "Deploying BPEL deployment %s", bpelDeployment);

    // now the bpel process will be parsed.
    final InputSource bpel = bpelDeployment.getBpel();
    if (bpel == null) {
      throw new OrchestraRuntimeException("Unable to open BPEL Stream");
    }

    final Parse parse = this.bpelParser.createParse();

    final WsdlDefinitions wsdlDefinitions = bpelDeployment.getWsdlDefinitions();
    parse.pushObject(wsdlDefinitions);

    final Namespaces namespaces = new Namespaces(parse);
    parse.pushObject(namespaces);

    ProcessFullDefinition processDefinition = null;
    final ProcessResourcesRepository resourcesRepository = bpelDeployment.getResourcesRepository();
    parse.pushObject(resourcesRepository);

    final WsdlsInfos wsdlInfos = new WsdlsInfos();
    parse.pushObject(wsdlInfos);

    BpelProcess bpelProcess = null;

    final Repository repository = environment.get(Repository.class);

    // make the bpel process available to the bpel parser by pushing it
    // on the object stack in the parse.
    final DefinitionKeeper definitionKeeper = new DefinitionKeeper(repository.getNextProcessCount());
    parse.pushObject(definitionKeeper);

    try {
      bpelProcess = (BpelProcess) parse.setInputSource(bpel).execute().getDocumentObject();
    } finally {
      // remove defKeeper
      parse.popObject();
      // remove wsdlInfos
      parse.popObject();
      // remove styleSheetRepository
      parse.popObject();
      // remove namespaces
      parse.popObject();
      // remove wsdlDefinitions
      parse.popObject();
    }
    if (!parse.hasProblems()) {
      if (bpelProcess == null) {
        throw new OrchestraRuntimeException("Bpel Process null");
      }
      bpelProcess.staticAnalysisCheck(parse);
    }
    parse.checkProblems("bpel file");

    final QName processQName = bpelProcess.getQName();
    final BpelProcess oldBpelProcess = repository.getProcess(processQName);
    if (oldBpelProcess != null) {
      throw new DeploymentException("Problem during deployment : process " + processQName + " already deployed");
    }
    repository.storeProcess(bpelProcess);

    // Record process deployed
    processDefinition = definitionKeeper.getProcessFullDefinition();
    EnvTool.getRecorder().recordProcessDeployed(processDefinition);

    Misc.log(Level.INFO, "BPEL process %s deployed.", processQName);
    return processDefinition;
  }

  public void undeploy(final QName processQName, final Environment environment) {
    final Repository repository = environment.get(Repository.class);
    final BpelProcess bpelProcess = repository.removeProcess(processQName);

    this.undeploy(bpelProcess.getUUID(), environment);

    Misc.log(Level.INFO, "BPEL process %s undeployed.", processQName);
  }

  public void undeploy(final ProcessDefinitionUUID pduuid, final Environment environment) {
    final Repository repository = environment.get(Repository.class);
    final BpelProcess bpelProcess = repository.removeProcess(pduuid);
    if (bpelProcess == null) {
      throw new DeploymentException("Problem during undeployment : process " + pduuid + " not found");
    }
    final List<BpelExecution> runningInstances = repository.getInstances(bpelProcess.getQName());
    if (runningInstances != null && !runningInstances.isEmpty()) {
      throw new DeploymentException(
          "Problem during undeployment: some instances  of process " + pduuid + " are still running");
    }
    EnvTool.getRecorder().recordProcessUndeployed(bpelProcess.getUUID());

    final ProcessDefinition processDefinition = EnvTool.getQuerier().getProcessDefinition(bpelProcess.getUUID());
    final UndeployedProcessHandler handler = EnvTool.get(UndeployedProcessHandler.class, UndeployedProcessHandler.DEFAULT_KEY);
    handler.handleUndeployedProcess(processDefinition);
    Misc.log(Level.INFO, "BPEL process %s undeployed.", pduuid);

  }
}
