/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.jmx;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import javax.xml.namespace.QName;

import org.ow2.orchestra.cluster.ClusterDescription;
import org.ow2.orchestra.cluster.LocalServer;
import org.ow2.orchestra.cluster.Server;
import org.ow2.orchestra.facade.Deployment;
import org.ow2.orchestra.facade.ManagementAPI;
import org.ow2.orchestra.facade.data.runtime.JobData;
import org.ow2.orchestra.facade.def.ActivityDefinition;
import org.ow2.orchestra.facade.def.ProcessDefinition;
import org.ow2.orchestra.facade.def.ProcessState;
import org.ow2.orchestra.facade.exception.ActivityInstanceNotFoundException;
import org.ow2.orchestra.facade.exception.ActivityNotFoundException;
import org.ow2.orchestra.facade.exception.InstanceNotFoundException;
import org.ow2.orchestra.facade.exception.JobNotFoundException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.exception.OrchestraWrapperException;
import org.ow2.orchestra.facade.exception.ProcessNotFoundException;
import org.ow2.orchestra.facade.jmx.RemoteDeployerMBean;
import org.ow2.orchestra.facade.runtime.ActivityInstance;
import org.ow2.orchestra.facade.runtime.ActivityState;
import org.ow2.orchestra.facade.runtime.ProcessInstance;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.jmx.commands.DeployCommand;
import org.ow2.orchestra.jmx.commands.DeployWSCommand;
import org.ow2.orchestra.jmx.commands.ExitInstanceCommand;
import org.ow2.orchestra.jmx.commands.FindActivityDefinitionsCommand;
import org.ow2.orchestra.jmx.commands.FindActivityInstancesCommand;
import org.ow2.orchestra.jmx.commands.FindJobsWithExceptionCommand;
import org.ow2.orchestra.jmx.commands.FindProcessDefinitionsCommand;
import org.ow2.orchestra.jmx.commands.FindProcessInstancesCommand;
import org.ow2.orchestra.jmx.commands.GetActivityDefinitionCommand;
import org.ow2.orchestra.jmx.commands.GetActivityInstanceCommand;
import org.ow2.orchestra.jmx.commands.GetDeployedProcessDefinitionUUIDCommand;
import org.ow2.orchestra.jmx.commands.GetProcessDefinitionCommand;
import org.ow2.orchestra.jmx.commands.GetProcessInstanceCommand;
import org.ow2.orchestra.jmx.commands.InitializeCommand;
import org.ow2.orchestra.jmx.commands.RemoveInstanceCommand;
import org.ow2.orchestra.jmx.commands.RemoveProcessCommand;
import org.ow2.orchestra.jmx.commands.ResumeInstanceCommand;
import org.ow2.orchestra.jmx.commands.RetryJobCommand;
import org.ow2.orchestra.jmx.commands.SuspendInstanceCommand;
import org.ow2.orchestra.jmx.commands.UndeployCommand;
import org.ow2.orchestra.jmx.commands.UndeployWSCommand;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.env.EnvironmentFactory;
import org.ow2.orchestra.pvm.internal.cmd.Command;
import org.ow2.orchestra.pvm.internal.cmd.CommandService;
import org.ow2.orchestra.util.DeploymentUtil;
import org.ow2.orchestra.util.Misc;

public class RemoteDeployerImpl implements RemoteDeployerMBean {

  private static Logger LOG = Logger.getLogger(RemoteDeployerImpl.class.getName());


  private final Server localServer;
  private final EnvironmentFactory environmentFactory;

  public RemoteDeployerImpl(final EnvironmentFactory environmentFactory) {
    this.localServer = new LocalServer(RemoteDeployerImpl.this);
    this.environmentFactory = environmentFactory;
  }


  public RemoteDeployerImpl(final Server orchestraServer, final EnvironmentFactory environmentFactory) {
    this.localServer = orchestraServer;
    this.environmentFactory = environmentFactory;
  }


  protected CommandService getCommandService() {
    return this.environmentFactory.get(CommandService.class);
  }

  public <T> T execute(final Command<T> command) {
    return this.getCommandService().execute(command);
  }

  public ProcessDefinition deploy(final Deployment bpelDeployment) {
    final ProcessDefinition processDefinition = this.execute(new DeployCommand(bpelDeployment));
    if (processDefinition != null) {
      try {
        this.deployWebServicesOnCluster(processDefinition.getUUID(), false);
      } catch (final RuntimeException e) {
        this.undeployWebServicesOnCluster(processDefinition.getUUID(), true);
        this.execute(new UndeployCommand(processDefinition.getUUID()));
        throw e;
      }
    }
    return processDefinition;
  }

  protected void deployWebServicesOnCluster(final ProcessDefinitionUUID processDefinitionUUID, final boolean ignoreExceptions) {
    // Get serverList
    final Collection<Server> servers = this.getOrchestraServersInCluster();
    // Deploy web services on each server
    for (final Server server : servers) {
      ManagementAPI managementAPI = null;
      if (server.equals(this.localServer)) {
        managementAPI = this;
      } else {
        // try to connect to server
        try {
          managementAPI = server.getManagementAPI();
        } catch (final Exception e) {
          // server is down, log a warning.
          RemoteDeployerImpl.LOG.log(Level.WARNING, "Connection to server " + server + " failed.", e);
        }
      }
      // If server is up
      if (managementAPI != null) {
        try {
          // deploy web services on the server
          managementAPI.deployWebServices(processDefinitionUUID);
          RemoteDeployerImpl.LOG.log(Level.INFO,
              "Web services for process " + processDefinitionUUID + " successfully deployed on server " + server + ".");
        } catch (final RuntimeException e) {
          // if exception
          if (ignoreExceptions) {
            RemoteDeployerImpl.LOG.log(Level.WARNING,
                "Web services deployment for process " + processDefinitionUUID + " failed on server " + server + ".", e);
          } else {
            throw e;
          }
        }
      }
    }
  }

  protected void undeployWebServicesOnCluster(final ProcessDefinitionUUID processDefinitionUUID, final boolean ignoreExceptions) {
    // Get serverList
    final Collection<Server> servers = this.getOrchestraServersInCluster();
    // Undeploy web services on each server
    for (final Server server : servers) {
      ManagementAPI managementAPI = null;
      // try to connect to server
      try {
        managementAPI = server.getManagementAPI();
      } catch (final Exception e) {
        // server is down, log a warning.
        RemoteDeployerImpl.LOG.log(Level.WARNING, "Connection to server " + server + " failed.", e);
      }
      // If server is up
      if (managementAPI != null) {
        try {
          // undeploy web services on the server
          managementAPI.undeployWebServices(processDefinitionUUID);
          RemoteDeployerImpl.LOG.log(Level.INFO,
              "Web services for process " + processDefinitionUUID + " successfully undeployed on server " + server + ".");
        } catch (final RuntimeException e) {
          // if exception
          if (ignoreExceptions) {
            RemoteDeployerImpl.LOG.log(Level.WARNING,
                "Web services undeployment for process " + processDefinitionUUID + " failed on server " + server + ".", e);
          } else {
            throw e;
          }
        }
      }
    }
  }

  private Collection<Server> getOrchestraServersInCluster() {
    return this.execute(new Command<Collection<Server>>() {

      public Collection<Server> execute(final Environment environment) {
        final ClusterDescription clusterDescription = environment.get(ClusterDescription.class);
        if (clusterDescription != null) {
          return clusterDescription.getOrchestraServers();
        }
        return Collections.singletonList(RemoteDeployerImpl.this.localServer);
      }
    });
  }


  public void deployWebServices(final ProcessDefinitionUUID processDefinitionUUID) {
    this.execute(new DeployWSCommand(processDefinitionUUID));
  }

  public void undeployWebServices(final ProcessDefinitionUUID processDefinitionUUID) {
    this.execute(new UndeployWSCommand(processDefinitionUUID));
  }

  public boolean undeploy(final QName processQName) {
    final ProcessDefinitionUUID processDefinitionUUID = this.execute(new GetDeployedProcessDefinitionUUIDCommand(processQName));
    return this.undeploy(processDefinitionUUID);
  }

  @SuppressWarnings("boxing")
  public boolean undeploy(final ProcessDefinitionUUID pduuid) {
    this.undeployWebServicesOnCluster(pduuid, false);

    try {
      return this.getCommandService().execute(new UndeployCommand(pduuid));
    } catch (final RuntimeException e) {
      this.deployWebServicesOnCluster(pduuid, true);
      throw e;
    }

  }

  public boolean remove(final ProcessDefinitionUUID processDefinitionUUID) {
    return this.getCommandService().execute(new RemoveProcessCommand(processDefinitionUUID));
  }

  public List<ProcessDefinition> findProcessDefinitions() {
    return this.getCommandService().execute(new FindProcessDefinitionsCommand());
  }

  public List<ProcessDefinition> findProcessDefinitions(final ProcessState state) {
    return this.getCommandService().execute(new FindProcessDefinitionsCommand(state));
  }

  public List<ProcessDefinition> findProcessDefinitions(final QName processName, final ProcessState state) {
    return this.getCommandService().execute(new FindProcessDefinitionsCommand(state, processName));
  }

  public List<ProcessDefinition> findProcessDefinitions(final QName processName) {
    return this.getCommandService().execute(new FindProcessDefinitionsCommand(processName));
  }

  public ProcessDefinition getProcessDefinition(final ProcessDefinitionUUID processUUID) throws ProcessNotFoundException {
    try {
      return this.getCommandService().execute(new GetProcessDefinitionCommand(processUUID));
    } catch (final OrchestraWrapperException e) {
      if (e.getCause() instanceof ProcessNotFoundException) {
        throw (ProcessNotFoundException) e.getCause();
      }
      throw e;
    }
  }

  public ActivityDefinition getActivityDefinition(final ActivityDefinitionUUID activityDefUUID) throws ActivityNotFoundException {
    try {
      return this.getCommandService().execute(new GetActivityDefinitionCommand(activityDefUUID));
    } catch (final OrchestraWrapperException e) {
      if (e.getCause() instanceof ActivityNotFoundException) {
        throw (ActivityNotFoundException) e.getCause();
      }
      throw e;
    }
  }

  public Set<ActivityDefinition> findActivityDefinitions(final ProcessDefinitionUUID processUUID)
    throws ProcessNotFoundException {
    try {
      return this.getCommandService().execute(new FindActivityDefinitionsCommand(processUUID));
    } catch (final OrchestraWrapperException e) {
      if (e.getCause() instanceof ProcessNotFoundException) {
        throw (ProcessNotFoundException) e.getCause();
      }
      throw e;
    }
  }


  public ProcessDefinition deployBar(final byte[] barFile) {
    try {
      // Create temp dir
      new File(System.getProperty("java.io.tmpdir")).mkdirs();
      final File tmpDir = File.createTempFile("orch", null, null);
      tmpDir.delete();
      if (!tmpDir.mkdirs()) {
        throw new IOException("Cannot create the temporary directory '" + tmpDir + "'.");
      }

      final List<URL> wsdlUrls = new ArrayList<URL>();
      URL bpelFile = null;

      // Unzip file
      final InputStream in = new ByteArrayInputStream(barFile);
      final ZipInputStream zipInput = new ZipInputStream(in);
      ZipEntry zipFile;

      final Map<String, byte[]> resources = new HashMap<String, byte[]>();

      while ((zipFile = zipInput.getNextEntry()) != null) {

        // write the file to the disk
        // Create file
        final File outFile = new File(tmpDir.getAbsolutePath() + File.separator + zipFile.getName());
        if (zipFile.isDirectory()) {
          outFile.mkdirs();
        } else {
          final byte[] content = Misc.getAllContentFrom(zipInput);
          resources.put(zipFile.getName(), content);
          Misc.write(content, outFile);// Copy file content

          if (zipFile.getName().endsWith(".wsdl")) {
            wsdlUrls.add(outFile.toURI().toURL());
          } else if (zipFile.getName().endsWith(".bpel")) {
            if (bpelFile != null) {
              throw new OrchestraRuntimeException("Bar contains more than one bpel file.");
            }
            bpelFile = outFile.toURI().toURL();
          }
        }
      }
      zipInput.close();
      in.close();

      final Deployment deployment = DeploymentUtil.generateDeployment(bpelFile, wsdlUrls);

      deployment.getResourcesRepository().getResources().putAll(resources);
      // Deploy the process
      final ProcessDefinition process = this.deploy(deployment);

      // Delete temp dir
      Misc.deleteDir(tmpDir);

      return process;
    } catch (final IOException e) {
      throw new OrchestraRuntimeException("Unable to deploy bar", e);
    }
  }

  public void initialize() {
    this.getCommandService().execute(new InitializeCommand());
  }

  public Set<ActivityInstance> findActivityInstances(final ProcessInstanceUUID processUUID) throws InstanceNotFoundException {
    try {
      return this.getCommandService().execute(new FindActivityInstancesCommand(processUUID));
    } catch (final OrchestraWrapperException e) {
      if (e.getCause() instanceof InstanceNotFoundException) {
        throw (InstanceNotFoundException) e.getCause();
      }
      throw e;
    }
  }

  public Set<ActivityInstance> findActivityInstances(final ProcessInstanceUUID processUUID, final ActivityState activityState)
    throws InstanceNotFoundException {
    try {
      return this.getCommandService().execute(new FindActivityInstancesCommand(processUUID, activityState));
    } catch (final OrchestraWrapperException e) {
      if (e.getCause() instanceof InstanceNotFoundException) {
        throw (InstanceNotFoundException) e.getCause();
      }
      throw e;
    }
  }

  public Set<ProcessInstance> findProcessInstances() {
    return this.getCommandService().execute(new FindProcessInstancesCommand());
  }

  public Set<ProcessInstance> findProcessInstances(final ProcessDefinitionUUID processDefinitionUUID)
    throws ProcessNotFoundException {
    try {
      return this.getCommandService().execute(new FindProcessInstancesCommand(processDefinitionUUID));
    } catch (final OrchestraWrapperException e) {
      if (e.getCause() instanceof ProcessNotFoundException) {
        throw (ProcessNotFoundException) e.getCause();
      }
      throw e;
    }
  }

  public Set<ProcessInstance> findProcessInstances(final ProcessDefinitionUUID processDefinitionUUID,
    final ActivityState processInstanceState) throws ProcessNotFoundException {
    try {
      return this.getCommandService().execute(new FindProcessInstancesCommand(processDefinitionUUID, processInstanceState));
    } catch (final OrchestraWrapperException e) {
      if (e.getCause() instanceof ProcessNotFoundException) {
        throw (ProcessNotFoundException) e.getCause();
      }
      throw e;
    }
  }

  public Set<ProcessInstance> findProcessInstances(final ActivityState processInstanceState) {
    return this.getCommandService().execute(new FindProcessInstancesCommand(processInstanceState));
  }

  public ProcessInstance getProcessInstance(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {
    try {
      return this.getCommandService().execute(new GetProcessInstanceCommand(processInstanceUUID));
    } catch (final OrchestraWrapperException e) {
      if (e.getCause() instanceof InstanceNotFoundException) {
        throw (InstanceNotFoundException) e.getCause();
      }
      throw e;
    }
  }

  public ActivityInstance getActivityInstance(final ActivityInstanceUUID activityInstanceUUID)
    throws ActivityInstanceNotFoundException {
    try {
      return this.getCommandService().execute(new GetActivityInstanceCommand(activityInstanceUUID));
    } catch (final OrchestraWrapperException e) {
      if (e.getCause() instanceof ActivityInstanceNotFoundException) {
        throw (ActivityInstanceNotFoundException) e.getCause();
      }
      throw e;
    }
  }

  /**
   * Exit the given instance.
   *
   * @param processInstanceUUID
   *          the Id of the instance to exit.
   * @throws InstanceNotFoundException
   *           thrown if the given instance to exit is not found.
   */
  public void exit(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {
    final InstanceNotFoundException infe = this.getCommandService().execute(new ExitInstanceCommand(processInstanceUUID));
    if (infe != null) {
      throw infe;
    }
  }

  /**
   * Resume a previously suspend instance of a process.
   *
   * @param processInstanceUUID
   *          the process instance UUID that identify the instance to resume.
   * @throws InstanceNotFoundException
   *           thrown if the given instance to resume is not found.
   */
  public void resume(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {
    final InstanceNotFoundException infe = this.getCommandService().execute(new ResumeInstanceCommand(processInstanceUUID));

    if (infe != null) {
      throw infe;
    }


  }

  /**
   * Suspend the given instance of a process.
   *
   * @param processInstanceUUID
   *          the process instance UUID that identify the instance to suspend.
   * @throws InstanceNotFoundException
   *           thrown if the given instance to suspend is not found.
   */
  public void suspend(final ProcessInstanceUUID processInstanceUUID) throws InstanceNotFoundException {

    final InstanceNotFoundException infe = this.getCommandService().execute(new SuspendInstanceCommand(processInstanceUUID));

    if (infe != null) {
      throw infe;
    }
  }

  /* (non-Javadoc)
   * @see org.ow2.orchestra.facade.InstanceManagementAPI#remove(org.ow2.orchestra.facade.uuid.ProcessInstanceUUID)
   */
  public boolean remove(final ProcessInstanceUUID processInstanceUUID) {
    return this.getCommandService().execute(new RemoveInstanceCommand(processInstanceUUID));
  }


  /**
   * Get the list of asynchronous executions for which all the tries have failed and which will not be executed any more.
   */
  public List<JobData> findJobsWithException() {
    return this.getCommandService().execute(new FindJobsWithExceptionCommand());
  }

  /**
   * Reschedules a job (asynchronous execution).
   * The job executor will try to execute the job at most retries times.
   * If the job fails for each retry, then the job executor will not try to execute it
   * unless this method is called again for this job.
   * @param jobData the job to retry.
   * @param retries the number of retries to set for the job.
   * @throws JobNotFoundException if job does not exist.
   */
  @SuppressWarnings("boxing")
  public void retryJob(final JobData jobData, final int retries) throws JobNotFoundException {
    if (!this.getCommandService().execute(new RetryJobCommand(retries, jobData))) {
      // if command returned false, job was not found
      // throw exception
      throw new JobNotFoundException(jobData);
    }
  }
}
