package org.ow2.orchestra.persistence.db;

import java.io.Serializable;
import java.util.Collection;
import java.util.List;
import java.util.logging.Level;

import javax.xml.namespace.QName;

import org.hibernate.FlushMode;
import org.hibernate.LockMode;
import org.hibernate.LockOptions;
import org.hibernate.Query;
import org.hibernate.ScrollableResults;
import org.hibernate.Session;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.BpelProcess.InstanceCreationLock;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.deployment.ProcessDeploymentCounter;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.persistence.RuntimeDbSession;
import org.ow2.orchestra.pvm.internal.hibernate.HibernatePvmDbSession;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.WaitingExecution;
import org.ow2.orchestra.runtime.WaitingExecution.WaitingExecutionIterator;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.PendingMessage;
import org.ow2.orchestra.util.Misc;

public class RuntimeDbSessionImpl extends
    HibernatePvmDbSession implements RuntimeDbSession {

  public RuntimeDbSessionImpl(final Session session) {
    this.setSession(session);
    session.setFlushMode(FlushMode.COMMIT);
  }

  public BpelProcess findBpelProcess(final String namespace, final String name, final int version) {
    final Query query = this.getSession().getNamedQuery("findBpelProcess");
    query.setString("namespace", namespace);
    query.setString("name", name);
    query.setInteger("version", version);
    query.setCacheable(true);
    final BpelProcess processDefinition = (BpelProcess) query.uniqueResult();
    return processDefinition;
  }

  public BpelProcess findBpelProcess(final String namespace, final String name) {
    final Query query = this.getSession().getNamedQuery("findLatestBpelProcess");
    query.setString("namespace", namespace);
    query.setString("name", name);
    query.setCacheable(true);
    return (BpelProcess) query.uniqueResult();
  }

  public BpelProcess findBpelProcess(final ProcessDefinitionUUID pduuid, final boolean lock) {
    if (lock) {
      final Query query = this.getSession().getNamedQuery("findBpelProcessForLock");
      query.setString("uuid", pduuid.toString());
      query.setLockOptions(new LockOptions(LockMode.PESSIMISTIC_WRITE));
      // don't use cache for this request or the lock
      query.setCacheable(false);
      final InstanceCreationLock res = (InstanceCreationLock) query.uniqueResult();
      if (res == null) {
        return null;
      }
    }
    return this.getBpelProcess(pduuid);
  }

  public BpelProcess getBpelProcess(final ProcessDefinitionUUID pduuid) {
    final Query query = this.getSession().getNamedQuery("getBpelProcessByUUID");
    query.setString("uuid", pduuid.toString());
    query.setCacheable(true);
    final BpelProcess processDefinition = (BpelProcess) query.uniqueResult();
    return processDefinition;
  }

  public List<BpelProcess> getBpelProcesses() {
    final Query query = this.getSession().getNamedQuery("getBpelProcesses");
    query.setCacheable(true);
    return query.list();
  }

  private static class WaitingExecutionIteratorImpl implements WaitingExecutionIterator {

    private final ScrollableResults results;
    private final Session session;
    private WaitingExecution current = null;

    public WaitingExecutionIteratorImpl(final ScrollableResults results, final Session session) {
      this.results = results;
      this.session = session;
    }
    public WaitingExecution next() {
      if (this.current != null) {
        this.session.evict(this.current);
      }
      if (this.results.next()) {
        final Serializable id = (Serializable) this.results.get(0);
        this.current = (WaitingExecution) this.session.get(WaitingExecution.class, id);
        if (this.current == null) {
          Misc.log(Level.INFO, "WaitingExecution %s not found, skipping.", id);
          return this.next();
        }
        return this.current;
      }
      return null;
    }
  }
  public WaitingExecutionIterator getWaitingExecutions(final InboundMessageElement inboundMessageElement) {
    final Query query = this.getSession().getNamedQuery("getWaitingExecutions");
    query.setParameter("activity", inboundMessageElement);
    final ScrollableResults waitingExecutions = query.scroll();
    return new WaitingExecutionIteratorImpl(waitingExecutions, this.getSession());
  }

  public WaitingExecution getFirstWaitingExecution(final InboundMessageElement inboundMessageElement) {
    final Query query = this.getSession().getNamedQuery("getWaitingExecutions");
    query.setParameter("activity", inboundMessageElement);
    query.setMaxResults(1);
    final Serializable id = (Serializable) query.uniqueResult();
    if (id != null) {
      return this.get(WaitingExecution.class, id);
    }
    return null;
  }

  public Collection<PendingMessage> getPendingMessages(final OperationKey operationKey) {
    final Query query = this.getSession().getNamedQuery("getPendingMessages");
    query.setString("operationKey", operationKey.toString());
    return query.list();
  }

  public List<BpelExecution> getBpelInstances() {
    final Query query = this.getSession().getNamedQuery("getBpelInstances");
    return query.list();
  }

  public BpelExecution getBpelInstance(final ProcessInstanceUUID processInstanceUUID) {
    final Query query = this.getSession().getNamedQuery("getBpelInstance");
    query.setString("processInstanceUUID", processInstanceUUID.toString());
    final BpelExecution bpelExecution = (BpelExecution) query.uniqueResult();
    return bpelExecution;
  }

  public List<BpelExecution> getBpelInstances(final QName processQName) {
    final Query query = this.getSession().getNamedQuery("getBpelInstancesOfProcess");
    query.setString("namespace", processQName.getNamespaceURI());
    query.setString("name", processQName.getLocalPart());
    return query.list();
  }

  public ProcessDeploymentCounter getProcessDeploymentCounter() {
    return (ProcessDeploymentCounter) this.getSession().createCriteria(ProcessDeploymentCounter.class).uniqueResult();
  }
}
