/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.BaseCatchActivity;
import org.ow2.orchestra.definition.activity.Compensate;
import org.ow2.orchestra.definition.activity.IAbstractActivity;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.facade.data.runtime.ProcessInstanceData;
import org.ow2.orchestra.facade.runtime.ActivityState;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.runtime.ScopeState;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.pvm.Execution;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.internal.job.JobImpl;
import org.ow2.orchestra.pvm.internal.job.TimerImpl;
import org.ow2.orchestra.pvm.internal.jobexecutor.JobDbSession;
import org.ow2.orchestra.pvm.internal.model.ExecutionImpl;
import org.ow2.orchestra.pvm.job.Job;
import org.ow2.orchestra.pvm.job.Timer;
import org.ow2.orchestra.pvm.model.Node;
import org.ow2.orchestra.pvm.model.OpenExecution;
import org.ow2.orchestra.pvm.session.MessageSession;
import org.ow2.orchestra.services.CheckHasMessageJob;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.ReplierKey;
import org.ow2.orchestra.services.itf.Replier;
import org.ow2.orchestra.services.job.BpelTimerImpl;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.wsdl.PartnerLinkType;
import org.ow2.orchestra.wsdl.Property;
import org.ow2.orchestra.wsdl.PropertyAlias;
import org.ow2.orchestra.wsdl.WsdlsInfos;

/**
 * a runtime path of execution.
 *
 * @author Tom Baeyens, Charles Souillard
 */
public class BpelExecution extends ExecutionImpl {

  public static final String ADDITIONNAL_SCOPE_VARIABLES = "additionnalScopeVariables";
  public static final String SCOPE_STATE_VARIABLE = "scopeState";

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  protected ProcessInstanceUUID processInstanceUUID;

  protected ActivityInstanceUUID activityInstanceUUID;

  protected Date startedDate;

  /** log. */
  private static Logger log = Logger.getLogger(BpelExecution.class.getName());

  protected FlowRuntime flowRuntime;

  protected ForEachRuntime forEachRuntime;

  protected ScopeRuntime scopeRuntime;

  protected Set<Replier> repliers = new HashSet<Replier>();

  protected WaitingExecution waitingExecution;

  /**
   * State of the process instance.
   */
  protected ActivityState instanceState = null;

  /** State of the execution before the suspend. Used to restore the same state when resuming execution. */
  protected String resumeState;

  public BpelExecution() {
    this.id = UUID.randomUUID().toString();
  }

  @Override
  protected BpelExecution newChildExecution() {
    final BpelExecution child = new BpelExecution();
    child.setProcessInstanceUUID(this.processInstanceUUID);
    child.setActivityInstanceUUID(this.activityInstanceUUID);
    child.setStartedDate(this.startedDate);
    EnvTool.getRepository().storeExecution(child);
    return child;
  }

  @Override
  public BpelExecution getParent() {
    return (BpelExecution) super.getParent();
  }

  /**
   * Returns the rootExecution (instance).
   *
   * @return the rootExecution (instance)
   */
  @Override
  public BpelExecution getProcessInstance() {
    return (BpelExecution) super.getProcessInstance();
  }

  /**
   * Search for the flowRuntime having a link declared with the given name. We
   * start searching into parent and not into this execution as links used in an
   * activity (the current node) must be declared before (in an enclosing flow).
   * So if the current node is flow, we don't want to search into its links. If
   * this parent execution flowRuntime doesn't have one, look into parent
   *
   * @param linkName
   *          linkName
   * @return the flowRuntime having a link declared with the given name.
   */
  public FlowRuntime getFlowRuntimeWithLink(final String linkName) {
    final FlowRuntime fr = this.getFlowRuntime();
    if (fr != null && fr.hasLink(linkName)) {
      return fr;
    }
    if (this.getParent() != null) {
      return (this.getParent()).getFlowRuntimeWithLink(linkName);
    }
    return null;
  }

  public ScopeRuntime getCurrentScopeRuntime() {
    if (this.scopeRuntime != null) {
      return this.scopeRuntime;
    }
    if (this.parent != null) {
      return (this.getParent()).getCurrentScopeRuntime();
    }
    return null;
  }

  public ScopeRuntime getScopeRuntimeWithVariable(final String variableName) {
    if (this.scopeRuntime != null) {
      if (this.scopeRuntime.containsVariable(variableName)) {
        return this.scopeRuntime;
      }
    }
    if (this.parent != null) {
      return (this.getParent()).getScopeRuntimeWithVariable(variableName);
    }
    return null;
  }

  public ScopeRuntime getScopeRuntimeWithCS(final String csName) {
    if (this.scopeRuntime != null) {
      if (this.scopeRuntime.containsCorrelationSet(csName)) {
        return this.scopeRuntime;
      }
    }
    if (this.parent != null) {
      return (this.getParent()).getScopeRuntimeWithCS(csName);
    }
    return null;
  }

  public ScopeRuntime getScopeRuntimeWithPLNK(final String partnerLinkName) {
    if (this.scopeRuntime != null) {
      if (this.scopeRuntime.getPartnerLink(partnerLinkName) != null) {
        return this.scopeRuntime;
      }
    }
    if (this.parent != null) {
      return (this.getParent()).getScopeRuntimeWithPLNK(partnerLinkName);
    }
    return null;
  }

  public ScopeRuntime getScopeRuntimeWithScope(final IAbstractActivity scope) {
    if (this.scopeRuntime != null) {
      if (this.scopeRuntime.getScopeDefinition().equals(scope)) {
        return this.scopeRuntime;
      }
    }
    if (this.parent != null) {
      return (this.getParent()).getScopeRuntimeWithScope(scope);
    }
    return null;
  }

  /**
   * Set the given linkStatus to the first enclosing flow having a link with the
   * given linkName.
   *
   * @param linkName
   *          linkName of the searched link
   * @param linkStatus
   *          linkStatus to set
   */
  public void setLinkStatus(final String linkName, final Boolean linkStatus) {
    final FlowRuntime fr = this.getFlowRuntimeWithLink(linkName);
    fr.setLinkStatus(linkName, linkStatus);
  }

  /**
   * toString method.
   *
   * @return String string representation of this execution.
   */
  @Override
  public String toString() {
    final StringBuffer sb = new StringBuffer("BpelExecution");

    if (this.node != null) {
      final IAbstractActivity activity = (IAbstractActivity) this.node.getBehaviour();
      sb.append('[');
      sb.append(activity.getName());
      sb.append(']');
    }

    sb.append(this.hashCode());
    return sb.toString();
  }

  /**
   * The process of this execution is guaranteed to be a {@link BpelProcess}.
   */
  @Override
  public BpelProcess getProcessDefinition() {
    return (BpelProcess) this.processDefinition;
  }

  /**
   * Search in currentScope (and parents) for a partnerLinkRuntime having the
   * given name.
   *
   * @param partnerLinkName
   *          partnerLinkName of the searched partnerLinkRuntime
   * @return the partnerLinkRuntime having the given partnerLinkName in the
   *         closest enclosing scope.
   */
  public PartnerLinkRuntime getPartnerLinkRuntime(final String partnerLinkName) {
    final ScopeRuntime scopeRuntime = this.getScopeRuntimeWithPLNK(partnerLinkName);
    if (scopeRuntime != null) {
      return scopeRuntime.getPartnerLink(partnerLinkName);
    }
    return null;
  }

  /**
   *
   * /** Search for a variableRuntime having the given variableName in the
   * closest scope and return its value.
   *
   * @param variableName
   *          variableName
   * @return Search for a variableRuntime having the given variableName in the
   *         closest scope and return its value.
   */
  public Object getVariableValue(final String variableName) {
    final VariableRuntime variable = this.getVariableByName(variableName);
    return variable.getValue();
  }

  /**
   * Search for a variableRuntime having the given variableName in the closest
   * scope.
   *
   * @param variableName
   *          variableName
   * @return the found variableRuntime
   */
  public VariableRuntime getVariableByName(final String variableName) {
    if (BpelExecution.log.isLoggable(Level.FINE)) {
      BpelExecution.log.fine("Getting variable " + variableName);
    }
    final ScopeRuntime currentRuntimeScope = this.getScopeRuntimeWithVariable(variableName);
    if (currentRuntimeScope != null) {
      return currentRuntimeScope.getVariable(variableName);
    }
    return null;
  }

  /**
   * Search for a correlationSetRuntime having the given csName in the closest
   * scope.
   *
   * @param csName
   *          csName
   * @return the found correlationSetRuntime
   */
  public CorrelationSetRuntime getCorrelationSetByName(final String csName) {
    if (BpelExecution.log.isLoggable(Level.FINE)) {
      BpelExecution.log.fine("Getting CorrelationSet " + csName);
    }
    final ScopeRuntime scopeRuntime = this.getScopeRuntimeWithCS(csName);
    if (scopeRuntime != null) {
      return scopeRuntime.getCorrelationSet(csName);
    }
    return null;
  }

  public PartnerLinkType getPartnerLinkType(final QName partnerLinkTypeName) {
    return this.getWsdlInfos().getPartnerLinkType(partnerLinkTypeName);
  }

  public Property getProperty(final QName propertyQName) {
    return this.getWsdlInfos().getProperty(propertyQName);
  }

  public PropertyAlias getPropertyAliasWithMessageType(final QName propertyName, final QName messageType) {
    return this.getWsdlInfos().getPropertyAliasWithMessageType(propertyName, messageType);
  }

  public List<PropertyAlias> getPropertyAliasesWithMessageType(final QName messageType) {
    return this.getWsdlInfos().getPropertyAliasesWithMessageType(messageType);
  }

  public List<PropertyAlias> getPropertyAliases(final CorrelationSet cs, final QName messageType) {
    return this.getWsdlInfos().getPropertyAliases(cs.getProperties().values(), messageType);
  }

  public PropertyAlias getPropertyAliasWithElement(final QName propertyName, final QName element) {
    return this.getWsdlInfos().getPropertyAliasWithElement(propertyName, element);
  }

  public PropertyAlias getPropertyAliasWithType(final QName propertyName, final QName type) {
    return this.getWsdlInfos().getPropertyAliasWithType(propertyName, type);
  }

  private WsdlsInfos getWsdlInfos() {
    return this.getProcessDefinition().getWsdlInfos();
  }

  public void pushFlowRuntime(final FlowRuntime flowRuntime) {
    if (this.flowRuntime == null) {
      this.flowRuntime = flowRuntime;
    } else {
      Misc.unreachableStatement("One execution should have only one flowRuntime runtime");
    }
  }

  public FlowRuntime getFlowRuntime() {
    return this.flowRuntime;
  }

  /**
   * @param forEachRuntime
   *          the forEachRuntime to push
   */
  public void pushForEachRuntime(final ForEachRuntime forEachRuntime) {
    if (this.forEachRuntime != null) {
      Misc.unreachableStatement("One execution should have only one foreach runtime");
    }
    this.forEachRuntime = forEachRuntime;
  }

  /**
   * Get the forEachRuntime.
   */
  public ForEachRuntime getForEachRuntime() {
    return this.forEachRuntime;
  }

  /**
   * @param scopeRuntime
   *          the scopeRuntime to push
   */
  public void pushScopeRuntime(final ScopeRuntime scopeRuntime) {
    if (this.scopeRuntime != null) {
      Misc.unreachableStatement("One execution should have only one scopeRuntime");
    }
    this.scopeRuntime = scopeRuntime;
  }

  /**
   * Pop a scopeRuntime.
   */
  public ScopeRuntime popScopeRuntime(final boolean shouldAdd) {
    if (this.scopeRuntime == null) {
      Misc.unreachableStatement("There is no scopeRuntime to pop!");
    }
    final ScopeRuntime res = this.scopeRuntime;
    this.scopeRuntime = null;
    if (shouldAdd) {
      // add this scopeRuntime as a child of parent scope runtime
      final ScopeRuntime parent = this.getCurrentScopeRuntime();
      if (parent == null) {
        Misc.unreachableStatement("parent is null");
      }
      parent.addEndedChild(res);
    }
    return res;
  }

  /**
   * Get the scopeRuntime.
   */
  public ScopeRuntime getScopeRuntime() {
    return this.scopeRuntime;
  }

  public Replier getReplier(final OperationKey operationKey, final String messageExchange) {
    if (this.scopeRuntime != null) {
      final ReplierKey replierKey = new ReplierKey(operationKey, messageExchange, this.scopeRuntime);
      final Iterator<Replier> replierIterator =  this.getRepliers().iterator();
      while (replierIterator.hasNext()) {
        final Replier replier = replierIterator.next();
        if (replier.getReplierKey().equals(replierKey)) {
          replierIterator.remove();
          return replier;
        }
      }
    }
    if (this.parent != null) {
      return this.getParent().getReplier(operationKey, messageExchange);
    }
    return null;
  }

  /**
   * @return the repliers
   */
  public Set<Replier> getRepliers() {
    if (this.getParent() != null) {
      return this.getProcessInstance().getRepliers();
    }
    return this.repliers;
  }

  public void addReplier(final Replier replier) {
    this.getRepliers().add(replier);
  }

  public void setWaitingExecution(final WaitingExecution waitingExecution) {
    this.waitingExecution = waitingExecution;
  }

  public void resetWaitingExecution() {
    this.waitingExecution = null;
  }

  public WaitingExecution getWaitingExecution() {
    return this.waitingExecution;
  }

  public boolean hasWaitingExecution() {
    return this.waitingExecution != null;
  }

  @Override
  public int hashCode() {
    return this.id.hashCode();
  }

  @Override
  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof BpelExecution)) {
      return obj.equals(this);
    }
    final BpelExecution other = (BpelExecution) obj;
    return other.getId().equals(this.id);
  }


  public Set<BpelExecution> searchWaitingExecutions() {
    final Set<BpelExecution> res = new HashSet<BpelExecution>();
    if (this.hasWaitingExecution()) {
      res.add(this);
    }
    if (this.getExecutions() != null) {
      for (final OpenExecution child : this.getExecutions()) {
        res.addAll(((BpelExecution) child).searchWaitingExecutions());
      }
    }
    return res;
  }

  @Override
  // TODO: if timers can be created by using TimerDefinition objects, this
  // method should be removed.
  public void createTimer(final String eventName, final String signalName, final String dueDateDescription, final Date dueDate,
    final String repeat, final Boolean isExclusive, final Integer retries) {
    super.createTimer(eventName, signalName, dueDateDescription, dueDate, repeat, isExclusive, retries);
  }

  @Override
  protected TimerImpl instantiateTimer() {
    return new BpelTimerImpl();
  }

  /**
   * Method called to cleanly terminate child activities. If a child activity is
   * a scope, then invokes the termination handler.
   */
  public void terminate() {
    final ActivityType activityType = ((IAbstractActivity) this.node.getBehaviour()).getType();
    BpelExecution.log.finest("Terminating BpelExecution " + this + ", activity type " + activityType + ", node name "
      + this.node.getName());

    if (this.scopeRuntime != null && !ActivityType.SCOPE.equals(activityType)) {
      Misc.unreachableStatement("The execution " + this + " has a scopeRuntime but is not a " + ActivityType.SCOPE + " (it is a "
        + activityType + ")");
    }
    if (this.scopeRuntime == null && ActivityType.SCOPE.equals(activityType)) {
      Misc.unreachableStatement("The execution " + this + " doen't have any scopeRuntime but is a " + activityType);
    }
    if (this.scopeRuntime != null) {
      BpelExecution.log.finest("Scope state for BpelExecution " + this + " is " + this.scopeRuntime.getState());
    }

    final Collection<OpenExecution> childExecutions = this.getExecutions();
    if (childExecutions != null && !childExecutions.isEmpty()) {
      // If there's a child, terminate just one child and then return.
      final BpelExecution childBpelExecution = (BpelExecution) childExecutions.iterator().next();
      BpelExecution.log.finest("Terminating child " + childBpelExecution + " out of " + childExecutions.size() + " children");
      childBpelExecution.terminate();
      return;
    } else if (ActivityType.SCOPE.equals(activityType)
        && ScopeState.HANDLING_FAULT.equals(this.scopeRuntime.getState())) {
      // All enclosed activities have been terminated,
      // execute catch

      // Get to the Scope NodeImpl and all children
      final List<Node> nodes = this.node.getNodes();
      if (nodes == null) {
        Misc.unreachableStatement("The parent scope node doesn't have any children");
      }
      if (nodes.size() < Scope.BIGGEST_ACTIVITY_INDEX) {
        Misc.unreachableStatement("The parent scope node has only " + nodes.size() + " children, it should have had at least "
          + Scope.BIGGEST_ACTIVITY_INDEX);
      }

      Node faultHandler = BaseCatchActivity.findFaultHandler(this.scopeRuntime, nodes, this.scopeRuntime.getLastException());
      if (faultHandler == null) {
        // If no appropriate fault handler matched, use the catchAll handler
        // (wsbpel-v2.0-OS, page 129, parts 1 and 2, last 2 cases for both)
        faultHandler = nodes.get(Scope.CATCHALL_ACTIVITY);
      }

      BpelExecution.log.finest("Will execute the fault handler " + faultHandler + " in BpelExecution "
        + this);
      this.createExecution().execute(faultHandler);
    } else if (ActivityType.SCOPE.equals(activityType)
        && (ScopeState.RUNNING.equals(this.scopeRuntime.getState())
            || ScopeState.RETHROWING_FAULT.equals(this.scopeRuntime.getState()))) {
      // if there is no more child in a scope, execute its termination handler
      // then return.
      // the termination handler will call terminate again if we are in a nested
      // scope.
      this.scopeRuntime.setState(ScopeState.TERMINATING);

      // Get to the Scope NodeImpl and all children
      final List<Node> nodes = this.node.getNodes();
      if (nodes == null) {
        Misc.unreachableStatement("The scope node " + this.node + " doesn't have any children");
      }
      if (nodes.size() < Scope.BIGGEST_ACTIVITY_INDEX) {
        Misc.unreachableStatement("The scope node " + this.node + " has only " + nodes.size()
          + " children, it should have had at least " + Scope.BIGGEST_ACTIVITY_INDEX);
      }

      final Node terminationHandler = nodes.get(Scope.TERMINATION_ACTIVITY);
      final ExecutionImpl terminationHandlerExecution = this.createExecution();

      BpelExecution.log.finest("Will execute the termination handler " + terminationHandler + " in BpelExecution "
        + terminationHandlerExecution + " for the BpelExecution " + this);
      terminationHandlerExecution.execute(terminationHandler);
      return;
    } else if (ActivityType.FOR_EACH.equals(activityType) && this.getForEachRuntime() != null
      && this.getForEachRuntime().isCompleted()) {
      // for each:
      // if forEach is completed (not terminated because of a fault),
      // all remaining branches have been terminated
      // continue normal execution.
      this.signal("all-finished");
    } else if (ActivityType.COMPENSATION_HANDLER.equals(activityType) && this.getParent().getScopeRuntime() != null) {
      // compensation handler terminated
      if (this.getParent().getScopeRuntime().getLastException() != null) {
        final BpelExecution chExecution = this.getParent();
        // remove finished execution from parent
        chExecution.removeExecution(this);
        // end ch execution
        chExecution.end(Execution.STATE_ENDED);

        // remove scopeRuntime
        final ScopeRuntime scopeRuntime = chExecution.popScopeRuntime(false);
        scopeRuntime.setState(ScopeState.COMPENSATED);
        final BpelExecution compensateExecution = scopeRuntime.getCompensateExecution();
        scopeRuntime.setCompensateExecution(null);

        // remove finished execution from parent
        final BpelExecution parentExecution = chExecution.getParent();
        parentExecution.removeExecution(chExecution);

        // Signal calling compensate activity
        final Map<String,Object> params = new HashMap<String, Object>();
        params.put(Compensate.EXCEPTION_PARAMETER, scopeRuntime.getLastException());
        compensateExecution.signal(params);
      }
    } else if (ActivityType.SCOPE_EVENT_HANDLER.equals(activityType)) {
      // if there is no more child in another activity, just exit the activity
      // and propagate to parent.
      // Note that the propagation is not called when:
      // - in the scope handling a fault
      // - in the forEach activity that is successfully completed
      // This seems to be the only activities which causes termination to
      // called.
      // DO NOT RECORD termination, because this is only a virtual activity (not monitored).
      this.removeExecution();
      if (this.getState().equals(Execution.STATE_ASYNC)) {
        this.unlock();
        this.end(Execution.STATE_ENDED);
      } else {
        this.end(Execution.STATE_CANCELLED);
      }
      if (this.parent != null) {
        BpelExecution.log.finest("Terminating parent " + this.parent);
        ((BpelExecution) this.parent).terminate();
      }
    } else {
      // if there is no more child in another activity, just exit the activity
      // and propagate to parent.
      // Note that the propagation is not called when:
      // - in the scope handling a fault
      // - in the forEach activity that is successfully completed
      // This seems to be the only activities which causes termination to
      // called.
      this.removeExecution();
      if (this.getState().equals(Execution.STATE_ASYNC)) {
        this.unlock();
        this.end(Execution.STATE_ENDED);
      } else {
        this.end(Execution.STATE_CANCELLED);
        if (this.parent != null) {
          if (this.parent.getNode() != this.getNode()) {
            final IAbstractActivity activity = (IAbstractActivity) this.getNode().getBehaviour();
            EnvTool.getRecorder().recordActivityTerminated(activity.createRuntimeData(this));
          }
        }
      }
      if (this.parent != null) {
        BpelExecution.log.finest("Terminating parent " + this.parent);
        ((BpelExecution) this.parent).terminate();
      }
    }
  }

  /**
   * Method called to abruptly stop all child activities. If a child activity is
   * a scope, then the termination handler is NOT invoked.
   */
  public void exit() {
    if (this.isSuspended()) {
      this.setState(this.resumeState);
    }
    // Remove execution first to avoid getting called twice
    this.removeExecution();

    // Exit all children
    final Collection<OpenExecution> childExecutions = this.getExecutions();
    if (childExecutions != null) {
      for (final OpenExecution childExecution : new ArrayList<OpenExecution>(childExecutions)) {
        final BpelExecution childBpelExecution = (BpelExecution) childExecution;
        childBpelExecution.exit();
      }
    }
    if (this.getState().equals(Execution.STATE_ASYNC)) {
      this.unlock();
      this.end(Execution.STATE_ENDED);
    } else if (!this.isFinished()) {
      this.end(Execution.STATE_CANCELLED);
      if (this.getParent() != null && this.getParent().getNode() != this.getNode()) {
        final IAbstractActivity activity = (IAbstractActivity) this.getNode().getBehaviour();
        EnvTool.getRecorder().recordActivityExited(activity.createRuntimeData(this));
      }
    }
  }

  private void removeExecution() {
    // Delete execution from all receivers
    this.resetWaitingExecution();
    // remove jobs
    this.removeJobs();

    if (this.parent != null) {
      // Make sure this execution doesn't call back its parent
      this.setPropagation(Propagation.EXPLICIT);
      this.parent.removeExecution(this);
    } else {
      // This is the root BpelExecution object of the process instance
      this.setProcessInstanceState(ActivityState.EXITED);
    }
  }

  public ProcessInstanceUUID getProcessInstanceUUID() {
    return this.processInstanceUUID;
  }

  public void setProcessInstanceUUID(final ProcessInstanceUUID processInstanceUUID) {
    this.processInstanceUUID = processInstanceUUID;
  }

  /**
   * @return the activityInstanceUUID
   */
  public ActivityInstanceUUID getActivityInstanceUUID() {
    return this.activityInstanceUUID;
  }

  /**
   * @return the activityInstanceUUID
   */
  public ActivityInstanceUUID getEncloserActivityInstanceUUID() {
    return this.getEncloserActivityInstanceUUID(this.getActivityInstanceUUID());
  }

  /**
   * @return the activityInstanceUUID
   */
  private ActivityInstanceUUID getEncloserActivityInstanceUUID(final ActivityInstanceUUID enclosedActivityUUID) {
    final ActivityInstanceUUID activityInstanceUUID = this.getActivityInstanceUUID();
    final BpelExecution parentBpelExecution = this.getParent();
    if (enclosedActivityUUID.equals(activityInstanceUUID)) {
      if (parentBpelExecution == null) {
        return null;
      } else {
        return parentBpelExecution.getEncloserActivityInstanceUUID(enclosedActivityUUID);
      }
    } else {
      return activityInstanceUUID;
    }
  }

  /**
   * @param activityInstanceUUID
   *          the activityInstanceUUID to set
   */
  public void setActivityInstanceUUID(final ActivityInstanceUUID activityInstanceUUID) {
    this.activityInstanceUUID = activityInstanceUUID;
  }

  /**
   * @return the startedDate
   */
  public Date getStartedDate() {
    return this.startedDate;
  }

  /**
   * @param startedDate
   *          the startedDate to set
   */
  public void setStartedDate(final Date startedDate) {
    this.startedDate = startedDate;
  }

  public ProcessInstanceData getProcessInstanceData() {
    return new ProcessInstanceData(this.getProcessDefinition().getProcessDefinitionData(), this.getProcessInstanceUUID(),
        this.getProcessInstanceState());
  }

  /**
   * Gets the state of the instance.
   *
   * @return whether the instance is suspended or not.
   */
  public boolean isInstanceSuspended() {
    return ActivityState.SUSPENDED.equals(this.getProcessInstanceState());
  }

  /**
   * Suspends this instance.
   */
  public void suspendInstance() {

    if (ActivityState.RUNNING.equals(this.getProcessInstanceState())) {
      this.suspend((BpelExecution) this.processInstance);
      this.setProcessInstanceState(ActivityState.SUSPENDED);
      EnvTool.getRecorder().recordInstanceUpdated(this.getProcessInstanceData());
    }

  }

  /**
   * Resumes this instance.
   */
  public void resumeInstance() {

    if (ActivityState.SUSPENDED.equals(this.getProcessInstanceState())) {
      this.resume((BpelExecution) this.processInstance);
      this.setProcessInstanceState(ActivityState.RUNNING);
      EnvTool.getRecorder().recordInstanceUpdated(this.getProcessInstanceData());

//      Environment.getCurrent().get(MessageSession.class).send(
//        new AsyncFindExecToSignal(this.getProcessInstance(), this.getProcessInstance().getProcessDefinition().getUUID()));
    }

  }

  /**
   * Recursively resumes all terminal children of the instance.
   *
   * @param node
   *          the node on which the children have to be suspended.
   */
  private void resume(final BpelExecution node) {

    if (node.getResumeState().equals(Execution.STATE_ACTIVE)) {
      final IAbstractActivity activity = (IAbstractActivity) node.getNode().getBehaviour();
      // check for the Alarms in a Pick Activity. If an alarm is reached, other
      // onMessage activities must be removed from their associated receiver.
      if (activity.getType().equals(ActivityType.PICK)) {

        boolean removeOnMessage = false;
        final Iterator<JobImpl<?>> jobs = node.getJobs().iterator();

        // check if an alarm is already reached
        while (jobs.hasNext() && !removeOnMessage) {
          final Job j = jobs.next();
          if (j instanceof Timer) {
            // if the job is a timer, compare it with the current time.
            if (j.getDueDate().before(new Date(System.currentTimeMillis()))) {
              removeOnMessage = true;
            }
          } else if (j instanceof CheckHasMessageJob) {
            jobs.remove();
          }
        }

        if (removeOnMessage) {
          // remove all OnMessage for this Pick activity from the waiting list.
          node.resetWaitingExecution();
        } else {
          Environment.getFromCurrent(MessageSession.class).send(new CheckHasMessageJob(node));
        }
      } else if (activity.isInboundMessageElement()) {
        node.removeJobs();
        Environment.getFromCurrent(MessageSession.class).send(new CheckHasMessageJob(node));
      }
    }
    // do the resume
    for (final OpenExecution n : node.getExecutions()) {
      this.resume((BpelExecution) n);
    }
    if (node.isSuspended()) {
      this.resumeJobs(node);
      if (node.isSuspended()) {
        node.setState(node.getResumeState());
      }
      Logger.getLogger(this.getClass().getName()).log(Level.FINE, "Node " + node + " now running.");
    } else {
      Logger.getLogger(this.getClass().getName()).log(Level.FINE, "Node " + node + " is already running.");
    }
  }

  /**
   * Recursively suspends all terminal children of the instance.
   *
   * @param node
   *          the node on which the children have to be suspended.
   */
  private void suspend(final BpelExecution node) {

    if (!node.getExecutions().isEmpty()) {

      for (final OpenExecution n : node.getExecutions()) {
        this.suspend((BpelExecution) n);
      }

    }
    if (!node.isSuspended()) {
      node.setResumeState(node.getState());
      node.setState(Execution.STATE_SUSPENDED);
      this.suspendJobs(node);
      Logger.getLogger(this.getClass().getName()).log(Level.FINE, "Node " + node + " suspended.");
    } else {
      Logger.getLogger(this.getClass().getName()).log(Level.FINE,
          "Node " + node + " is already suspended.");
    }
  }


  /**
   * Suspends all the jobs of a given node.
   *
   * @param node
   *          the node to suspend the jobs on.
   */
  private void suspendJobs(final OpenExecution node) {
    for (final Job j : node.getJobs()) {
      final JobImpl<?> job = (JobImpl) j;
      job.setSuspended(true);
      if (job.getLockOwner() != null) {
        job.setLockOwner(null);
        job.setLockExpirationTime(null);
      }
    }
  }

  /**
   * Resumes all the jobs of a given node.
   *
   * @param node
   *          the node to resume the jobs on.
   */
  private void resumeJobs(final OpenExecution node) {
    for (final Job j : node.getJobs()) {
      final JobImpl<?> job = (JobImpl) j;
      job.setSuspended(false);
      if (job.getLockOwner() != null) {
        job.setLockOwner(null);
        job.setLockExpirationTime(null);
      }
    }
  }

  public void removeJob(final Job job) {
    Environment.getFromCurrent(JobDbSession.class).delete(job);
  }

  public void removeJobs() {
    for (final Job job : new HashSet<Job>(this.getJobs())) {
      Environment.getFromCurrent(JobDbSession.class).delete(job);
    }
  }

  public ActivityState getProcessInstanceState() {
    if (this.getParent() == null) {
      return this.instanceState;
    } else {
      return this.getProcessInstance().getProcessInstanceState();
    }
  }

  public void setProcessInstanceState(final ActivityState state) {
    if (this.getParent() == null) {
      this.instanceState = state;
    } else {
      this.getProcessInstance().setProcessInstanceState(state);
    }
  }


  /**
   * @return the resumeState
   */
  public String getResumeState() {
    return this.resumeState;
  }


  /**
   * @param resumeState the resumeState to set
   */
  public void setResumeState(final String resumeState) {
    this.resumeState = resumeState;
  }

  @Override
  public void removeExecution(final Execution child) {
    super.removeExecution(child);
    // Delete execution from all receivers
    ((BpelExecution)child).resetWaitingExecution();
    // remove jobs
    ((BpelExecution)child).removeJobs();
    EnvTool.getRepository().removeExecution(child);
  }
}
