/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.util.logging.Logger;

import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.facade.data.runtime.PartnerLinkUpdateData;
import org.ow2.orchestra.facade.runtime.PartnerLinkUpdateType;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * runtime representation of one partner for a {@link PartnerLinkRole}.
 *
 * @author Tom Baeyens, Charles SOuillard
 */
public class PartnerLinkRuntime {

  /** log. */
  private static Logger log = Logger.getLogger(PartnerLinkRuntime.class
      .getName());

  protected long dbid;
  protected int dbversion;

  protected ActivityInstanceUUID scopeActivityInstanceUUID;

  /** name of this partnerLink. */
  protected String name;

  /** definition of this partnerLink. */
  protected PartnerLink partnerLinkDefinition;

  /**
   * The value of the endpointReference associated to the partnerRole of the
   * partnerLink. It is a mechanism that allows you to select one of the
   * available services in the WSDL, or even define new services at runtime.
   */
  private Document partnerRoleEndPointReference;

  /**
   * The value of the endpointReference associated to the myRole of the
   * partnerLink. It is a mechanism that allows you to select one of the
   * available services in the WSDL, or even define new services at runtime.
   */
  private Document myRoleEndPointReference;

  protected PartnerLinkRuntime() {
  }

  /**
   * Default constructor.
   *
   * @param partnerLinkDefinition
   *            partnerLinkDefinition of this partnerLinkRuntime.
   * @param execution
   *            execution to which this partnerLinkRuntime belongs to.
   */
  public PartnerLinkRuntime(final PartnerLink partnerLinkDefinition,
      final BpelExecution execution, final ActivityInstanceUUID scopeActivityInstanceUUID) {
    this.scopeActivityInstanceUUID = scopeActivityInstanceUUID;
    this.name = partnerLinkDefinition.getName();
    this.partnerLinkDefinition = partnerLinkDefinition;
    if (partnerLinkDefinition.getMyRolePortTypeQName() != null) {
      this.initializeMyRole(execution);
    }
    if (partnerLinkDefinition.getPartnerRolePortTypeQName() != null) {
      this.initializePartnerRole(execution);
    }
  }

  /**
   * Initialize partnerRole endpoint reference to the defaultPort value.
   *
   * @param execution
   *            execution to which this partnerLinkRuntime belongs to
   */
  public void initializePartnerRole(final BpelExecution execution) {
    if (this.partnerLinkDefinition.getDefaultPartnerRoleEndPointReference() != null) {
      EnvTool.getRecorder().recordPartnerLinkUpdated(
        new PartnerLinkUpdateData(execution.getActivityInstanceUUID(), execution.getActivityInstanceUUID(), this.name,
          PartnerLinkUpdateType.PARTNER_ROLE_INITIALIZED, this.partnerLinkDefinition.getDefaultPartnerRoleEndPointReference().getDocumentElement()));
    }
  }

  /**
   * Initialize myRole endpoint reference to the defaultPort value.
   *
   * @param execution
   *            execution to which this partnerLinkRuntime belongs to
   */
  public void initializeMyRole(final BpelExecution execution) {
    if (this.partnerLinkDefinition.getDefaultMyRoleEndPointReference() != null) {
      EnvTool.getRecorder().recordPartnerLinkUpdated(
        new PartnerLinkUpdateData(execution.getActivityInstanceUUID(), execution.getActivityInstanceUUID(), this.name,
          PartnerLinkUpdateType.MY_ROLE_INITIALIZED, this.partnerLinkDefinition.getDefaultMyRoleEndPointReference().getDocumentElement()));
    }
  }

  public Element getMyRoleEndPointReference() {
    if (this.myRoleEndPointReference != null) {
      return this.myRoleEndPointReference.getDocumentElement();
    }
    if (this.partnerLinkDefinition.getDefaultMyRoleEndPointReference() != null) {
      return this.partnerLinkDefinition.getDefaultMyRoleEndPointReference().getDocumentElement();
    }
    return null;
  }

  public void setMyRoleEndPointReference(final Element myRoleEndPointReference) {
    final Document doc = XmlUtil.getNewDocument();
    doc.appendChild(doc.importNode(myRoleEndPointReference, true));
    this.myRoleEndPointReference = doc;
  }

  public Element getPartnerRoleEndPointReference() {
    if (this.partnerRoleEndPointReference != null) {
      return this.partnerRoleEndPointReference.getDocumentElement();
    }
    if (this.partnerLinkDefinition.getDefaultPartnerRoleEndPointReference() != null) {
      return this.partnerLinkDefinition.getDefaultPartnerRoleEndPointReference().getDocumentElement();
    }
    return null;
  }

  public void setPartnerRoleEndPointReference(
      final Element partnerRoleEndPointReference,
      final ActivityInstanceUUID activityInstanceUUID) {
    final Document doc = XmlUtil.getNewDocument();
    doc.appendChild(doc.importNode(partnerRoleEndPointReference, true));
    this.partnerRoleEndPointReference = doc;
    EnvTool.getRecorder().recordPartnerLinkUpdated(
      new PartnerLinkUpdateData(this.scopeActivityInstanceUUID, activityInstanceUUID, this.name,
        PartnerLinkUpdateType.PARTNER_ROLE_UPDATED, this.partnerRoleEndPointReference.getDocumentElement()));
  }

  public String getName() {
    return this.name;
  }

  public PartnerLink getPartnerLinkDefinition() {
    return this.partnerLinkDefinition;
  }

  public long getDbid() {
    return this.dbid;
  }

}
