/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services;

import java.lang.reflect.Method;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import javax.transaction.Synchronization;
import javax.wsdl.Definition;

import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.env.Transaction;
import org.ow2.orchestra.pvm.internal.cmd.Command;
import org.ow2.orchestra.pvm.internal.cmd.CommandService;
import org.ow2.orchestra.pvm.internal.job.JobImpl;
import org.ow2.orchestra.pvm.internal.jobexecutor.JobExceptionHandler;
import org.ow2.orchestra.pvm.job.Job;
import org.ow2.orchestra.pvm.session.MessageSession;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.impl.PublisherImpl;
import org.ow2.orchestra.services.impl.PublisherImpl.ProcessEndpoint;
import org.ow2.orchestra.services.itf.Invoker;
import org.ow2.orchestra.services.job.ExecuteInvokeJob;
import org.ow2.orchestra.services.job.SignalInvokeJob;
import org.ow2.orchestra.util.AddressingUtil;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.OrchestraConstants;
import org.ow2.orchestra.util.PrefixThreadFactory;
import org.ow2.orchestra.util.AddressingUtil.AddressingInfo;
import org.ow2.orchestra.var.MessageVariable;
import org.w3c.dom.Element;


/**
 * @author Guillaume Porcher
 *
 */
public class InvokeExecutor {

  protected static class FinishInvokeCommand implements Command<Void> {

    private BpelExecution execution;

    private final long executionId;
    private final MessageVariable responseMessage;
    private final Exception exception;

    public FinishInvokeCommand(final long executionId, final MessageVariable responseMessage) {
      this.executionId = executionId;
      this.responseMessage = responseMessage;
      this.exception = null;
    }

    public FinishInvokeCommand(final long executionId, final Exception exception) {
      this.executionId = executionId;
      this.responseMessage = null;
      this.exception = exception;
    }


    /**
     * @param execution the execution to set
     */
    public void setExecution(final BpelExecution execution) {
      this.execution = execution;
    }

    public Void execute(final Environment environment) {
      final BpelExecution bpelExecution;
      if (this.execution == null) {
        // synchronize execution with session
        bpelExecution = EnvTool.getRepository().getUpdatedExecution(this.executionId);
      } else {
        bpelExecution = this.execution;
      }
      if (bpelExecution == null) {
        InvokeExecutor.LOG.info("Invoke response not delivered to execution "
            + this.executionId + ": execution has been removed.");
        return null;
      }
      if (bpelExecution.isFinished()) {
        InvokeExecutor.LOG.info("Invoke response not delivered to execution " + bpelExecution + ": execution is finished.");
        return null;
      }

      // look for ExecuteInvokeJob previously defined for this execution
      ExecuteInvokeJob invokeJob = null;
      final Iterator<JobImpl<?>> jobIterator = bpelExecution.getJobs().iterator();
      while (jobIterator.hasNext()) {
        final Job job = jobIterator.next();
        if (job instanceof ExecuteInvokeJob) {
          invokeJob = (ExecuteInvokeJob) job;
          break;
        }
      }
      if (invokeJob == null) {
        Misc.unreachableStatement("unable to find invoke job");
        return null;
      }
      if (this.exception == null || (this.exception instanceof BpelFaultException)) {
        // Invoke was successful => Signal execution
        InvokeExecutor.LOG.info("Deliver invoke response to execution " + bpelExecution);

        // remove invokeJob and save isSuspended value
        final boolean isSuspended = invokeJob.isSuspended();
        bpelExecution.removeJob(invokeJob);


        // Create SignalInvokeJob
        // set same suspended mode
        final SignalInvokeJob signalInvokeJob = new SignalInvokeJob(this.responseMessage, this.exception);
        signalInvokeJob.setExecution(bpelExecution);
        signalInvokeJob.setSuspended(isSuspended);

        // schedule job execution
        environment.get(MessageSession.class).send(signalInvokeJob);
      } else {
        // internal exception during invoke: reschedule invoke execution
        final JobExceptionHandler jobExceptionHandler =
          new JobExceptionHandler(invokeJob, this.exception);
        jobExceptionHandler.execute(environment);
      }
      return null;
    }

  }

  private static class InvokeCall implements Runnable {

    private final Invoker invoker;
    private final OperationKey operationKey;
    private final AddressingInfo addressingInfo;
    private final MessageVariable requestMessage;
    private final long executionId;
    private final CommandService commandService;
    private final Set<Definition> wsdlDefinitions;
    private final Properties orchestraProperties;

    public InvokeCall(
        final long executionId,
        final Invoker invoker,
        final OperationKey operationKey,
        final AddressingInfo addressingInfo,
        final MessageVariable requestMessage,
        final Set<Definition> wsdlDefinitions,
        final CommandService commandService,
        final Properties orchestraProperties) {
      this.executionId = executionId;
      this.invoker = invoker;
      this.operationKey = operationKey;
      this.addressingInfo = addressingInfo;
      this.requestMessage = requestMessage;
      this.wsdlDefinitions = wsdlDefinitions;
      this.commandService = commandService;
      this.orchestraProperties = orchestraProperties;
    }

    public void run() {
      FinishInvokeCommand fic;
      try {
        final MessageVariable responseMsg =
          this.invoker.invoke(
              this.operationKey,
              this.addressingInfo,
              this.requestMessage,
              this.wsdlDefinitions,
              this.orchestraProperties);
        fic = new FinishInvokeCommand(this.executionId, responseMsg);
      } catch (final Exception e) {
        fic = new FinishInvokeCommand(this.executionId, e);
      }
      this.commandService.execute(fic);
    }

  }

  private static final Logger LOG = Logger.getLogger(InvokeExecutor.class.getName());

  private final ThreadPoolExecutor executorService;

  public InvokeExecutor(final int maxThreads) {
    final ThreadPoolExecutor tpe = new ThreadPoolExecutor(maxThreads, maxThreads,
        60L, TimeUnit.SECONDS,
        new LinkedBlockingQueue<Runnable>(), new PrefixThreadFactory("InvokeExecutor"));
    try {
      final Method m = ThreadPoolExecutor.class.getDeclaredMethod("allowCoreThreadTimeOut", boolean.class);
      m.invoke(tpe, true);
    } catch (final Exception e) {
      // method is only available in java 6
    }
    this.executorService = tpe;
  }

  public void executeInvoke(
      final Environment environment,
      final BpelExecution execution,
      final OperationKey operationKey,
      final Element partnerEndPointReference,
      final MessageVariable requestMessage,
      final Set<Definition> wsdlDefinitions) {
    final CommandService commandService = environment.get(CommandService.class);
    final Properties orchestraProperties = (Properties) Environment.getFromCurrent(OrchestraConstants.ORCHESTRA_PROPERTIES);
    final long executionId = execution.getDbid();

    final AddressingInfo addressingInfo = AddressingUtil.getAddressingInfo(partnerEndPointReference, orchestraProperties);
    final PublisherImpl publisher = environment.get(PublisherImpl.class);
    if (publisher != null) {

      final ProcessEndpoint processEndpoint = publisher.getProcessEndpoint(addressingInfo.getAddress());

      if (processEndpoint != null) {
        // direct process call
        final OperationKey processOpKey = new OperationKey(
            processEndpoint.getProcessDefinitionUUID(),
            operationKey.getPortTypeQName(),
            operationKey.getOperationName());
        ProcessInstanceMessageCarrier interProcessMessageCarrier = null;
        final boolean isOneWay = processEndpoint.isOneWay(operationKey.getOperationName());

        if (!isOneWay) {
          interProcessMessageCarrier = new ProcessInstanceMessageCarrier(executionId);
        }
        /*
         * Call orchestra
         */
        ReceivingService.handle(
            requestMessage,
            processOpKey,
            interProcessMessageCarrier,
            processEndpoint.isLock(operationKey.getOperationName()),
            commandService,
            execution.getProcessInstance().getParentInstanceStartTime());
        if (isOneWay) {
          final FinishInvokeCommand fic = new FinishInvokeCommand(executionId, (MessageVariable) null);
          fic.setExecution(execution);

          fic.execute(environment);
        }
        return;
      }
    }


    final Invoker invoker = EnvTool.get(Invoker.class, Invoker.DEFAULT_KEY);
    final Transaction transaction = environment.get(Transaction.class);

    final InvokeCall invokeCall = new InvokeCall(
        executionId,
        invoker,
        operationKey,
        addressingInfo,
        requestMessage,
        wsdlDefinitions,
        commandService,
        orchestraProperties);
    if (transaction == null) {
      // we are outside a transaction, directly schedule task
      this.executorService.execute(invokeCall);
    } else {
      // we are in a transaction, schedule task after the end of the transaction
      transaction.registerSynchronization(
          new Synchronization() {
            public void afterCompletion(final int status) {
              if (status == javax.transaction.Status.STATUS_COMMITTED) {
                InvokeExecutor.this.executorService.execute(invokeCall);
              }
            }

            public void beforeCompletion() {
              // do nothing
            }
          }
      );
    }
  }


  /**
   * @return the executorService
   */
  public ThreadPoolExecutor getExecutorService() {
    return this.executorService;
  }
}
