/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services.impl;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;

import javax.wsdl.Definition;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.IAbstractActivity;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.internal.model.NodeImpl;
import org.ow2.orchestra.pvm.model.Node;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.util.Misc;

public class PublisherImpl implements Publisher {

  public static class ProcessEndpoint {
    private ProcessDefinitionUUID processDefinitionUUID;
    private Map<String, Boolean> isOneWay;
    private Map<String, Boolean> locks;

    public boolean isLock(final String operationName) {
      return this.locks.get(operationName);
    }

    public boolean isOneWay(final String operationName) {
      return this.isOneWay.get(operationName);
    }

    public ProcessDefinitionUUID getProcessDefinitionUUID() {
      return this.processDefinitionUUID;
    }
  }
  private final Map<ProcessDefinitionUUID, Set<Definition>> wsdlDefinitions =
    new HashMap<ProcessDefinitionUUID, Set<Definition>>();

  private final Map<String, ProcessEndpoint> processEndpoints =
    new HashMap<String, ProcessEndpoint>();

  public void publishServices(final BpelProcess bpelProcess, final Environment environment) {
    Misc.log(Level.INFO, "Publishing services for Bpel process %s (%s).", bpelProcess.getUUID(), bpelProcess.getQName());
    this.wsdlDefinitions.put(bpelProcess.getUUID(),
        new HashSet<Definition>(bpelProcess.getWsdlInfos().getWsdlDefinitions()));
  }

  public void unpublishServices(final BpelProcess bpelProcess, final Environment environment) {
    Misc.log(Level.INFO, "Unpublishing services for Bpel process %s (%s).", bpelProcess.getUUID(), bpelProcess.getQName());
    this.wsdlDefinitions.remove(bpelProcess.getUUID());
  }

  protected Set<OperationKey> getServiceOperationKeys(final BpelProcess bpelProcess, final NodeImpl node,
      final Environment environment) {

    final Set<OperationKey> operationKeys = new HashSet<OperationKey>();
    // get list of operation keys for the current node
    final IAbstractActivity behaviour = (IAbstractActivity) node.getBehaviour();
    if (behaviour.isInboundMessageElement()) {
      final InboundMessageElement inboundMessageElement = (InboundMessageElement) behaviour;
      for (final ReceivingElement receivingElement : inboundMessageElement.getReceivingElements()) {
        final OperationKey operationKey = receivingElement.getOperationKey();
        operationKeys.add(operationKey);
      }
    }
    // Add list of operation keys defined in children nodes
    final List<Node> children = node.getNodes();
    if (children != null) {
      for (final Node child : children) {
        operationKeys.addAll(this.getServiceOperationKeys(bpelProcess, (NodeImpl) child, environment));
      }
    }
    return operationKeys;
  }

  public Set<Definition> getWsdlDefinitions(final ProcessDefinitionUUID processDefinitionUUID) {
    return this.wsdlDefinitions.get(processDefinitionUUID);
  }

  public void addProcessEndpoint(final String endpoint,
      final ProcessDefinitionUUID processDefinitionUUID,
      final Map<String, Boolean> locks, final Map<String, Boolean> isOneWay) {
    final ProcessEndpoint processEndpoint = new ProcessEndpoint();
    processEndpoint.processDefinitionUUID = processDefinitionUUID;
    processEndpoint.locks = locks;
    processEndpoint.isOneWay = isOneWay;
    this.processEndpoints.put(endpoint, processEndpoint);
  }

  public void removeProcessEndpoint(final String endpoint) {
    this.processEndpoints.remove(endpoint);
  }

  public ProcessEndpoint getProcessEndpoint(final String endpoint) {
    return this.processEndpoints.get(endpoint);
  }
}
