/**
 * Copyright (C) 2007-2009  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services.job;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.Map;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.activity.Invoke;
import org.ow2.orchestra.exception.FaultWithMessageVariable;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.pvm.env.Environment;
import org.ow2.orchestra.pvm.internal.job.MessageImpl;
import org.ow2.orchestra.pvm.internal.jobexecutor.JobDbSession;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.var.MessageVariable;

/**
 * @author Guillaume Porcher, S. Ali Tokmen
 */
public class SignalInvokeJob extends MessageImpl<Void> {

  public static class SignalInvokeJobData {

    private MessageVariable responseMessage;
    private String exceptionMessage;
    private QName faultName;

    public SignalInvokeJobData() {
    }
    public MessageVariable getResponseMessage() {
      return this.responseMessage;
    }
    public void setResponseMessage(final MessageVariable responseMessage) {
      this.responseMessage = responseMessage;
    }
    public String getExceptionMessage() {
      return this.exceptionMessage;
    }
    public void setExceptionMessage(final String exceptionMessage) {
      this.exceptionMessage = exceptionMessage;
    }
    public QName getFaultName() {
      return this.faultName;
    }
    public void setFaultName(final QName faultName) {
      this.faultName = faultName;
    }
  }

  private static final long serialVersionUID = 1796627705334254482L;

  protected SignalInvokeJobData data = new SignalInvokeJobData();

  // for persistence
  protected SignalInvokeJob() {
  }

  public SignalInvokeJob(final MessageVariable responseMessage, final Exception exception) {
    this.data.setResponseMessage(responseMessage != null ? responseMessage.duplicate() : null);
    this.data.setExceptionMessage(null);
    if (exception != null) {
      if (exception instanceof FaultWithMessageVariable) {
        final FaultWithMessageVariable fault = (FaultWithMessageVariable) exception;
        this.data.setFaultName(fault.getFaultName());
        this.data.setResponseMessage(fault.getFaultMessage());
      } else {
        // IMPORTANT: We assume FaultWithMessage can only be created in internal
        // by Orchestra, and will never be returned by an invoked WS
        //
        // serialize the stack trace
        final StringWriter sw = new StringWriter();
        exception.printStackTrace(new PrintWriter(sw));
        this.data.setExceptionMessage(sw.toString());
      }
    }
  }

  /**
   * Signals execution waiting on an invoke activity. Provides received message
   * as parameter.
   */
  public Void execute(final Environment environment) {
    final BpelExecution execution = (BpelExecution) this.getExecution();

    final Map<String, Object> signalParams = new HashMap<String, Object>();

    if (this.data.getFaultName() != null) {
      signalParams.put(Invoke.RECEIVED_EXCEPTION, new FaultWithMessageVariable(this.data.getFaultName(), this.data.getResponseMessage()));
    } else if (this.data.getExceptionMessage() != null && !this.data.getExceptionMessage().equals("")) {
      signalParams.put(Invoke.RECEIVED_EXCEPTION, new OrchestraRuntimeException(this.data.getExceptionMessage()));
    } else {
      signalParams.put(Invoke.RECEIVED_MESSAGE, this.data.getResponseMessage());
    }

    execution.signal(signalParams);

    final JobDbSession jobDbSession = environment.get(JobDbSession.class);
    jobDbSession.delete(this);

    return null;
  }

}
